#!/usr/bin/perl -T
#
# darcs.cgi - the darcs repository viewer
#
# Copyright (c) 2004 Will Glozer
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#
# This program calls darcs (or its own subroutines) to generate XML
# which is rendered into HTML by XSLT.  It is capable of displaying
# the files in a repository, various patch histories, annotations, etc.
#

require strict;

use CGI qw( :standard );
use CGI::Util;
use File::Basename;
use File::stat;
use File::Temp;
use IO::Handle;
use POSIX;

## the following variables can be customized to reflect your system
## configuration by defining them appropriately in the file
## "/etc/cgi.conf".  The syntax accepts equals signs or simply
## blanks separating values from assignments.

$ENV{'PATH'} = read_conf('PATH', $ENV{'PATH'});

# path to executables, or just the executable if they are in $ENV{'PATH'}
$darcs_program    = read_conf("darcs", "darcs");
$xslt_program     = read_conf("xsltproc", "xsltproc");

# directory containing repositories
$repository_root  = read_conf("reposdir", "/var/www");

# XSLT template locations
$template_root = read_conf("xslt_dir", "/usr/local/share/darcs/xslt");

$xslt_annotate = "$template_root/annotate.xslt";
$xslt_browse   = "$template_root/browse.xslt";
$xslt_patches  = "$template_root/patches.xslt";
$xslt_repos    = "$template_root/repos.xslt";

$xslt_errors   = "$template_root/errors.xslt";

# CSS stylesheet that XSLT templates refer to.  This is a HTTP request
# path, not a local file system path. The default will magically read
# /etc/darcs/styles.css
$stylesheet    = read_conf("stylesheet", "/cgi-bin/darcs.cgi/styles.css");

# XML source for the error pages
$xml_errors = "$template_root/errors.xml";

## end customization

# ----------------------------------------------------------------------

# read a value from the cgi.conf file.

sub read_conf {
  my $flag, $val;
  if ($#_ == 0) {
    ($flag) = @_;
    $val = "";
  } else {
    ($flag, $val) = @_;
  }
  if (open(CGI_CONF, "/etc/darcs/cgi.conf")) {
    while (<CGI_CONF>) {
      if (/^\s*$flag\s*=\s*(\S+)/) {
        $val = $1;
      } elsif (/^\s*$flag\s+(\S+)/) {
        $val = $1;
      }
    }
    close(CGI_CONF);
  }
  return $val;
}

# transform and output `xml' with stylesheet file `xslt'
sub transform {
    my ($xml, $xslt, $args) = @_;

    print "Content-type: text/html\r\n\r\n";
    print `$xslt_program $args $xslt $xml`;
}

# create a temporary XML document with a root element and the repository path
sub make_xml {
    my ($repo, $dir, $file) = @_;
    my ($fh, $fname) = File::Temp::tempfile();
    my ($full_path, $path) = '/';

    printf $fh qq(<darcs repository="$repo" target="%s/%s%s">\n),
        $repo, ($dir ? "$dir/" : ''), ($file ? "$file" : '');

    print $fh qq(<path>\n);
    foreach $path (split('/', "$repo/$dir")) {
        $full_path .= "$path/";
        print $fh qq(<directory full-path="$full_path">$path</directory>\n);
    }
    if ($file) {
        print $fh qq(<file full-path="$full_path$file">$file</file>\n) if $file;
    }
    print $fh qq(</path>\n\n);
    
    return ($fh, $fname);
}

# finish XML output
sub finish_xml {
    my ($fh) = @_;
    print $fh "\n</darcs>\n";
    $fh->flush;
}

# run darcs and collect the output in an XML document
sub darcs_xml {
    my ($repo, $cmd, $args, $dir, $file) = @_;
    my ($fh, $darcs_out) = make_xml($repo, $dir, $file);

    $args .= " --xml-output ";
    darcs($fh, $repo, $cmd, $args, $dir, $file);

    finish_xml($fh);
    return ($fh, $darcs_out);
}

# run darcs with output redirected to the specified file handle
sub darcs {
    my ($fh, $repo, $cmd, $args, $dir, $file) = @_;
    my ($target);

    # quote target only if there is one, otherwise darcs will get an empty param
    if ($dir || $file) {
        $target = sprintf("%s%s%s", $dir, ($dir ? '/' : ''), $file);
    }
    
    chdir "$repository_root/$repo";
    open(DARCS, "$darcs_program '$cmd' $args $target|");
    while (<DARCS>) {
        print $fh $_;
    }
    close(DARCS);
}

# get a directory listing as XML output
sub dir_listing {
    my ($repo, $dir) = @_;
    my ($fh, $dir_out) = make_xml($repo, $dir, '');
        
    print $fh "<files>\n";
    foreach $file (glob("$repository_root/$repo/_darcs/current/$dir/*")) {
        my $secs  = stat($file)->mtime;
        my $mtime = localtime($secs);
        my $ts = POSIX::strftime("%Y%m%d%H%M%S", gmtime $secs);
        
        my ($name, $type);

         if (-d $file) {
             ($name, $type) = (basename($file) . '/', 'directory');
         } else {
             ($name, $type) = (basename($file), 'file');            
         }
         printf $fh qq(  <$type name="$name" modified="$mtime" ts="$ts" />\n);
    }
    print $fh "</files>\n";

    finish_xml($fh);
    return ($fh, $dir_out);
}

# get a repository listing as XML output
sub repo_listing {
    my ($fh, $fname) = make_xml("", "", "");
        
    print $fh "<repositories>\n";
    foreach $repo (glob("$repository_root/*")) {
        if (-d "$repo/_darcs") {
            $name = basename($repo);
            printf $fh qq(  <repository name="$name" />\n);
        }
    }
    print $fh "</repositories>\n";

    finish_xml($fh);
    return ($fh, $fname);
}

# safely extract a parameter from the http request.  This applies a regexp
# to the parameter which should group only the appropriate parameter value
sub safe_param {
    my ($param, $regex, $default) = @_;
    my $value = CGI::Util::unescape(param($param));
    return ($value =~ $regex) ? $1 : $default;
}

# respond to a CGI request
sub respond {
    # untaint script_name, reasonable to expect only \w, -, /, and . in the name
    my $script_name = CGI::Util::unescape(script_name());
    $script_name =~ qr~^([\w/.\-]+)$~ or die qq(bad script_name "$script_name");
    $script_name = $1;

    # untaint simple parameters, which can only have chars matching \w+
    my $cmd  = safe_param('c', '^(\w+)$', 'browse');
    my $sort = safe_param('s', '^(\w+)$', '');

    # set the xslt processing arguments
    my $xslt_args = qq {
        --stringparam cgi-program '$script_name'
        --stringparam sort-by '$sort'
        --stringparam stylesheet '$stylesheet'
    };
    $xslt_args =~ s/\s+/ /gm;    

    my ($path) = CGI::Util::unescape(path_info());
    # don't allow ./ or ../ in paths
    $path =~ s|[.]+/||g;
    
    # check whether we're asking for styles.css
    if ($path eq '/styles.css') {
        open (STYLES_CSS, "/Users/will/temp/darcs/cgi/xslt/styles.css") or        
            die "couldn't find styles.css"; # FIXME should give default styles.css?
        print "Content-type: text/css\r\n\r\n";
        while (<STYLES_CSS>) {
          print $_;
        }
        close (STYLES_CSS);
        return;
    }

    # when no repository is requested display available repositories
    if (length($path) < 2) {
        ($fh, $fname) = repo_listing();
        transform($fname, $xslt_repos, $xslt_args);
        File::Temp::unlink0($fn, $fname);
        return;
    }

    # don't allow any shell meta characters in paths
    $path =~ qr@^([^\\!\$\^&*()\[\]{}<>~`|';"?\r\n]+)$@ or \
        die qq(bad path_info "$path");
    @path = split('/', substr($1, 1));

    # split the path into a repository, directory, and file
    my ($repo, $dir, $file, @bits) = '', '', '', [];
    while (@path > 0) {
        $repo = join('/', @path);
        # check if remaining path elements refer to a repo
        if (-d "${repository_root}/${repo}/_darcs") {
            if (@bits > 1) {
                $dir  = join('/', @bits[0..$#bits - 1]);
            }
            $file = $bits[$#bits];
            # check if last element of path, stored in $file, is really a dir
            if (-d "${repository_root}/${repo}/_darcs/current/${dir}/${file}") {
                $dir = ($dir ? "$dir/$file" : $file);
                $file = '';
            }
            last;
        } else {
            $repo = '';
            unshift(@bits, pop @path);
        }
    }
    
    # make sure the repository exists
    unless ($repo) {
        $xslt_args .= " --stringparam error-type 'invalid-repository' ";    
        print "Status: 404 Invalid repository\r\n";
        transform($xml_errors, $xslt_errors, $xslt_args);
        return;
    }
    
    # untaint patches and tags. Tags can have arbitrary values, so
    # never pass these unquoted, on pain of pain!
    my $patch = safe_param('p', '^([\w\-.]+)$');
    my $tag   = safe_param('t', '^(.+)$');    
        
    my $darcs_args = "";
    $darcs_args .= " --match 'hash $patch' " if $patch;
    $darcs_args .= " -t '$tag' " if $tag;    

    my ($fn, $fname);

    # process the requested command
    if ($cmd eq 'browse') {
        ($fh, $fname) = dir_listing($repo, $dir);
        transform($fname, $xslt_browse, $xslt_args);
    } elsif ($cmd eq 'patches') {
        ($fh, $fname) = darcs_xml($repo, "changes", $darcs_args, $dir, $file);
        transform($fname, $xslt_patches, $xslt_args);       
    } elsif ($cmd eq 'annotate') {
        $darcs_args .= " --summary ";
        ($fh, $fname) = darcs_xml($repo, "annotate", $darcs_args, $dir, $file);
        transform($fname, $xslt_annotate, $xslt_args);       
    } elsif ($cmd eq 'diff') {
        $darcs_args .= " -u ";
        print "Content-type: text/plain\r\n\r\n";
        darcs(STDOUT, $repo, "diff", $darcs_args, $dir, $file);
    } else {
        $xslt_args .= " --stringparam error-type 'invalid-command' ";
        print "Status: 400 Invalid command\r\n";
        transform($xml_errors, $xslt_errors, $xslt_args);
    }
    File::Temp::unlink0($fn, $fname);
}

# CGI entry point
respond();
